#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil
import subprocess
import tempfile

from datetime import datetime

import click
import jinja2

from click_plugins import with_plugins
from pkg_resources import iter_entry_points
from sagemaker.processing import ProcessingInput, ProcessingOutput
from sagemaker.sklearn.processing import SKLearnProcessor

from ..aws.events import (
    delete_event_rule,
    schedule_lambda_function,
    update_event_rule,
)
from ..aws.lambda_utils import (
    create_iam_role_for_lambda,
    create_lambda_deployment_package,
    delete_lambda_function,
    deploy_lambda_function,
    exponential_retry,
)
from ..aws.processing import get_sklearn_processor
from ..aws.s3 import download_s3_directory, get_s3_resource, parse_s3_path
from ..aws.session import get_boto_session
from ..config import read_yaml
from ..constants import (
    CONDA_ENV_NAME,
    EXCLUDE_FILES,
    EXCLUDE_FOLDERS,
    SAGEMAKER_CONTAINER_PATH_MAIN,
)
from ..log import get_logger, verbosity_option
from ..utils import render_template
from ..utils.zip import zip_directory
from . import rudderlabs

logger = get_logger(__name__)


def prepare_scripts(template_dir: str, context: dict, templates: list) -> str:
    """Prepare scripts for processing

    Args:
        template_dir: Path to the template directory
        context: Context for the template
        templates: List of template files to use
    Returns:
        str: Path to the scripts zip file
    """


def prepare_input_data(
    repository_path: str, input_path: str
) -> (ProcessingInput, bool):
    """Prepare input data for processing

    Args:
        repository_path: Path to the repository
        input_path: Path to the input data
    Returns:
        ProcessingInput: Input data for processing
        bool: Whether the input data is S3 or not
    """

    if input_path is None or input_path == "":
        return None, False

    sagemaker_input_path = os.path.join(SAGEMAKER_CONTAINER_PATH_MAIN, "data")

    logger.info(f"Input data is {input_path}")

    # s3 Location
    if input_path.startswith("s3://"):
        return (
            ProcessingInput(
                source=input_path,
                s3_data_type="S3Prefix",
                s3_data_distribution_type="FullyReplicated",
                destination=sagemaker_input_path,
            ),
            True,
        )

    # Local path
    # check input data path weather it is relative to repository or absolute
    if not os.path.isabs(input_path):
        input_path = os.path.join(repository_path, input_path)

    if not os.path.exists(input_path):
        return None, False

    logger.info("Preparing input data")
    input_data_zip_path = zip_directory(input_path)

    return (
        ProcessingInput(
            source=input_data_zip_path,
            destination=sagemaker_input_path,
        ),
        False,
    )


def prepare_output_data(
    output_data_path: str, output_name: str, running_mode: str = "remote"
) -> (ProcessingOutput, bool):
    """Prepare output data for processing

    Args:
        output_data_path: Path to the output data
        output_name: Name of the output data folder
        running_mode: Running mode of the pipeline (local or remote)

    Returns:
        ProcessingOutput: Output data for processing
        bool: Whether the output data is S3 or not
    """

    sagemaker_output_data_path = os.path.join(
        SAGEMAKER_CONTAINER_PATH_MAIN, "output"
    )

    logger.info(f"Output data is {output_data_path}")

    # s3 Location
    if output_data_path.startswith("s3://"):
        return (
            ProcessingOutput(
                source=sagemaker_output_data_path,
                output_name=output_name,
                destination=output_data_path,
                s3_upload_mode="EndOfJob"
                if running_mode == "local"
                else "Continuous",
            ),
            True,
        )
    else:
        return (
            ProcessingOutput(
                source=sagemaker_output_data_path, output_name=output_name
            ),
            False,
        )


def download_output_data(
    sklearn_processor: SKLearnProcessor,
    running_mode: str,
    local_output_path: str,
    creds: dict,
    job_id: str,
    output_name: str,
) -> None:
    """Download output data from S3 to local directory

    Args:
        sklearn_processor: SKLearnProcessor object
        running_mode: Running mode of the pipeline (local or remote)
        local_output_path: Local path to download output data
        creds: Credentials for S3
        job_id: Job ID of the pipeline
        output_name: Name of the processing output

    Returns:
        None:
    """
    s3_bucket = None
    process_job_output_path = None

    if running_mode != "local":
        s3_bucket = creds["s3Bucket"]
        process_job_output_path = (
            f"{sklearn_processor.latest_job.job_name}/output/output-1/{job_id}"
        )
    else:
        preprocessing_job_description = sklearn_processor.jobs[-1].describe()
        output_config = preprocessing_job_description["ProcessingOutputConfig"]

        for output in output_config["Outputs"]:
            if output["OutputName"] != output_name:
                continue

            output_s3_url = output["S3Output"]["S3Uri"]
            s3_bucket, process_job_output_path = parse_s3_path(output_s3_url)

    logger.info(f"S3 bucket: {s3_bucket}")
    logger.info(f"Process job output path: {process_job_output_path}")

    if s3_bucket is not None and process_job_output_path is not None:
        # Downloading model output files into local
        s3_resource = get_s3_resource(creds)

        download_s3_directory(
            s3_resource, s3_bucket, process_job_output_path, local_output_path
        )


def run_pipeline_step(
    pipeline_step_info: dict,
    creds: dict,
    instance_type: str,
    job_id: str,
    repository_path: str,
    exclude_folders: list,
    exclude_files: list,
) -> None:
    """Runs given pipeline step in sci-kit learn processor in amazon sagemaker

    Args:
        pipeline_step_info: Pipeline step information
        creds: AWS credentials
        instance_type: Instance type to use for the sagemaker job
        job_id: all the outputs will be saved under the folder with this name
        repository_path: Path to the repository
        exclude_folders: List of directories to be excluded from the zip
        exclude_files: List of files to be excluded from the zip

    Returns:
        None: None
    """
    # Get sklearn processor
    logger.info(f"Pipeline step: {pipeline_step_info['name']}")
    job_name = (
        f"{pipeline_step_info['name']}-{pipeline_step_info['job_suffix']}"
    )
    sklearn_processor = get_sklearn_processor(creds, instance_type, job_name)

    # Prepare source code and input data
    logger.info("Preparing source code")
    source_code_zip_path = zip_directory(
        repository_path,
        exclude_folders=exclude_folders,
        exclude_files=exclude_files,
    )

    sagemaker_code_path = os.path.join(SAGEMAKER_CONTAINER_PATH_MAIN, "code")
    sagemaker_req_path = os.path.join(
        SAGEMAKER_CONTAINER_PATH_MAIN, "requirements"
    )

    local_req_path = os.path.join(repository_path, "requirements.txt")

    # script parameters
    script_params = {f"{k}": v for k, v in pipeline_step_info["params"].items()}

    # Pass job id to the pipeline script as a parameter
    script_params["--job-id"] = job_id

    input_path = pipeline_step_info.get("input_path")
    input_path = (
        ""
        if input_path is None
        else input_path.replace("<job_id>", f"{job_id}")
    )

    # Prepare input data
    processing_input, input_location_is_s3 = prepare_input_data(
        repository_path, input_path
    )

    if not input_location_is_s3 and processing_input is not None:
        script_params["--input-data-zip"] = os.path.join(
            processing_input.destination,
            os.path.basename(processing_input.source),
        )

    if processing_input is not None:
        script_params["--input-data-path"] = processing_input.destination

    # Prepare output data
    output_path = pipeline_step_info.get("output_path")
    if output_path is None:
        raise ValueError("Output path is not specified")
        return

    output_path = output_path.replace("<job_id>", f"{job_id}")
    processing_output, output_location_is_s3 = prepare_output_data(
        output_path, pipeline_step_info["name"], instance_type
    )

    script_params["--output-data-path"] = processing_output.source

    script_params["--source-code-zip"] = os.path.join(
        sagemaker_code_path, os.path.basename(source_code_zip_path)
    )
    script_params["--requirements-path"] = os.path.join(
        sagemaker_req_path, "requirements.txt"
    )

    arguments = []
    for k, v in script_params.items():
        arguments.append(f"{k}")
        arguments.append(f"{v}")

    logger.info(f"Arguments: {arguments}")
    inputs = [
        ProcessingInput(
            source=source_code_zip_path, destination=sagemaker_code_path
        ),
        ProcessingInput(source=local_req_path, destination=sagemaker_req_path),
    ]

    if processing_input:
        inputs.append(processing_input)

    sklearn_processor.run(
        code=pipeline_step_info["code"],
        inputs=inputs,
        outputs=[processing_output],
        arguments=arguments,
    )

    if not output_location_is_s3:
        # Uploading model output files from s3 to local
        local_output_path = os.path.join(
            repository_path, pipeline_step_info["output_path"]
        )

        output_name = pipeline_step_info["name"]
        download_output_data(
            sklearn_processor,
            instance_type,
            local_output_path,
            creds,
            job_id,
            output_name,
        )

    # Cleaning up latest job files
    try:
        s3_bucket = sklearn_processor.sagemaker_session.default_bucket()
        root_prefix = sklearn_processor.latest_job.job_name

        logger.info(
            f"Cleaning S3 data located at bucket: {s3_bucket}, prefix :{root_prefix}"
        )

        s3_resource = get_s3_resource(creds)
        s3_resource.Bucket(s3_bucket).objects.filter(
            Prefix=root_prefix
        ).delete()
    except Exception as e:
        logger.error(f"Error deleting S3 object: {e}")


@click.command(
    epilog="""
    The command to run given notebookes in the pipeline.

    Examples:

        $ rlabs aws run-pipeline --pipeline-config-file pipeline.yaml --credentials-file credentials.yaml --repository-path /path/to/repository --instance-type ml.t3.xlarge --job-id my-job-id

        $ rlabs aws run-pipeline -p pipeline.yaml -c credentials.yaml -r /path/to/repository -i local -j my-job-id
    """
)
@click.option(
    "-j",
    "--job-id",
    default=None,
    help="Job id to be used for the pipeline, used to store output files in S3/local",
)
@click.option(
    "-c",
    "--credentials-file",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    show_default=True,
    default=os.path.join(os.path.realpath(os.curdir), "credentials.yaml"),
)
@click.option(
    "-i",
    "--instance-type",
    default="ml.t3.xlarge",
    show_default=True,
    help="The instance type to use for the amazon sagemaker notebook instance.",
)
@click.option(
    "-p",
    "--pipeline-config-file",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    help="The pipeline config file to use.",
)
@click.option(
    "-r",
    "--repository-path",
    default=os.path.realpath(os.curdir),
    show_default=True,
    type=click.Path(exists=True, readable=True, resolve_path=True),
    help="The repository path to use.",
)
@verbosity_option()
@rudderlabs.raise_on_error
def run(
    job_id: str,
    credentials_file: click.Path,
    instance_type: str,
    pipeline_config_file: click.Path,
    repository_path: click.Path,
) -> None:
    """Run the pipeline defined in the pipeline config file."""
    logger.info("Running pipeline")
    logger.info("credentials_file: %s", credentials_file)
    logger.info("Instance type: %s", instance_type)

    if job_id is None:
        job_id = int(datetime.now().timestamp())

    # Load the pipeline config file
    pipeline_config = read_yaml(pipeline_config_file)
    logger.info("Pipeline config: %s", pipeline_config)

    # Load the credentials file
    config = read_yaml(credentials_file)
    exclude_folders = (
        pipeline_config.get("exclude_folders", []) + EXCLUDE_FOLDERS
    )
    exclude_files = pipeline_config.get("exclude_files", []) + EXCLUDE_FILES

    # Runing pipeline
    for pipeline_step in pipeline_config["pipeline"]:
        logger.info("\n\n\nRUNNING PIPELINE STEP: %s", pipeline_step["name"])

        run_pipeline_step(
            pipeline_step_info=pipeline_step,
            creds=config,
            instance_type=instance_type,
            job_id=job_id,
            repository_path=repository_path,
            exclude_folders=exclude_folders,
            exclude_files=exclude_files,
        )


@click.command(epilog="""See rlabs aws pipeline --help for more information.""")
@click.option(
    "-id",
    "--instance-id",
    required=True,
    help="The instance id to use for scheduling the pipeline.",
)
@click.option(
    "-sit",
    "--sagemaker-instance-type",
    default="ml.t3.xlarge",
    show_default=True,
    help="The instance type to use for the amazon sagemaker notebook instance.",
)
@click.option(
    "-p",
    "--pem-file",
    required=True,
    type=click.Path(exists=True, resolve_path=True),
    help="The pem key file for communicating with the instance.",
)
@click.option(
    "-u",
    "--ec2-username",
    default="ec2-user",
    show_default=True,
    help="EC2 username for the instance.",
)
@click.option(
    "-r",
    "--repository-path",
    default=os.path.realpath(os.curdir),
    show_default=True,
    type=click.Path(exists=True, readable=True, resolve_path=True),
    help="The repository path to use.",
)
@click.option(
    "-pn",
    "--project-name",
    default=None,
    type=str,
    help="Name of the project, this will be used as a prefix for creating roles, lambda functions, event rules etc.",
)
@click.option(
    "-pc",
    "--pipeline-config-file",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    help="The pipeline config file to use. relative to the repository path.",
)
@click.option(
    "-c",
    "--credentials-file",
    type=click.Path(exists=True, readable=True, resolve_path=True),
    show_default=True,
    default=os.path.join(os.path.realpath(os.curdir), "credentials.yaml"),
    help="The credentials file to use. relative to the repository path.",
)
@click.option(
    "-es",
    "--event-schedule",
    default="rate(1 day)",
    show_default=True,
    help="The specified schedule in either cron or rate format. More info at https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-create-rule-schedule.html",
)
@click.option(
    "-rn",
    "--role-name",
    default=None,
    help="The name of the role to use for the lambda function. If not specified, a new role will be created with the name <repository_name>-lambda-role",
)
@verbosity_option()
@rudderlabs.raise_on_error
def schedule(
    instance_id: str,
    sagemaker_instance_type: str,
    pem_file: str,
    ec2_username: str,
    repository_path: str,
    project_name: str,
    pipeline_config_file: str,
    credentials_file: str,
    event_schedule: str,
    role_name: str,
) -> None:
    """
    Setup the instance and schedule the pipeline.
    """
    logger.info("Setting up instance: %s", instance_id)

    if sagemaker_instance_type == "local":
        logger.error("Local instance type is not supported")
        return

    # Load the credentials file
    creds = read_yaml(credentials_file)
    boto_session = get_boto_session(creds)
    ec2_resource = boto_session.resource("ec2")
    instance = ec2_resource.Instance(instance_id)

    # Adding trailing slash to repository path if not present
    # This can be used to find relative path of pipeline config file and credentials file
    # And can be useful while rsyncing to defferent project folder, it will be the case
    # when specifying project name is different from the repository directory name
    repository_path = os.path.join(repository_path, "")

    if instance.state["Name"] != "running":
        logger.info("Instance is not running, starting instance")
        instance.start()

        logger.info("Waiting for instance to start")
        instance.wait_until_running()

        logger.info("Waiting for status ok")
        waiter = ec2_resource.meta.client.get_waiter("instance_status_ok")
        waiter.wait(InstanceIds=[instance_id])

    credentials_path_relative = os.path.relpath(
        credentials_file, repository_path
    )

    if project_name is None:
        project_name = os.path.basename(repository_path[:-1])

    credentials_file_name = os.path.basename(credentials_file)

    # take relative path for pipeline config file
    pipeline_config_file_relative = pipeline_config_file.replace(
        repository_path, ""
    )

    # take relative path for credentials file
    credentials_file_relative = credentials_file.replace(repository_path, "")

    pipeline_run_command = f"rlabs aws pipeline run -p {pipeline_config_file_relative} -c {credentials_path_relative} -i {sagemaker_instance_type}"
    logger.info("Pipeline run command: %s", pipeline_run_command)

    # Jinja2 context for the template
    context = dict(
        user=ec2_username,
        project_name=project_name,
        conda_env=CONDA_ENV_NAME,
        project_dir=f"/home/{ec2_username}/{project_name}",
        pipeline_run_command=pipeline_run_command,
        instance_id=instance_id,
        credentials_file=credentials_path_relative,
        region=creds["aws"]["region"],
    )

    # base jinja2 engine
    template_loader = jinja2.PackageLoader(
        "rudderlabs.data.apps", os.path.join("templates", "scripts")
    )
    env = jinja2.Environment(
        loader=template_loader,
        autoescape=jinja2.select_autoescape(["html", "xml"]),
    )

    # Render the templates
    scripts_dir = tempfile.mkdtemp()
    accepted_extensions = [".sh", ".py"]
    for k in template_loader.list_templates():
        ext = os.path.splitext(k)[1]
        if ext not in accepted_extensions:
            continue
        render_template(env, k, context, scripts_dir)

    logger.info("Copying setup scripts to instance")
    host = f"{ec2_username}@{instance.public_ip_address}"
    host_scripts_dir = f"/home/{ec2_username}/scripts"

    logger.info("Syncing project repository")
    subprocess.run(
        [
            "rsync",
            "-rav",
            "-e",
            f"ssh -i {pem_file} -o StrictHostKeyChecking=no",
            "--exclude",
            "data",
            "--exclude",
            ".git",
            repository_path,
            f"{host}:/home/{ec2_username}/{project_name}",
        ]
    )

    subprocess.run(
        [
            "scp",
            "-o",
            "StrictHostKeyChecking=no",
            "-i",
            pem_file,
            "-r",
            scripts_dir,
            f"{host}:{host_scripts_dir}",
        ]
    )
    subprocess.run(
        [
            "scp",
            "-o",
            "StrictHostKeyChecking=no",
            "-i",
            pem_file,
            f"{scripts_dir}/run_pipeline.sh",
            f"{host}:/home/{ec2_username}/run_pipeline.sh",
        ]
    )

    logger.info("Creating conda environment")
    subprocess.run(
        [
            "ssh",
            "-i",
            pem_file,
            host,
            "sh",
            f"{host_scripts_dir}/create_conda_env.sh",
        ]
    )

    logger.info("Creating cron job")
    subprocess.run(
        [
            "ssh",
            "-i",
            pem_file,
            host,
            "sh",
            f"{host_scripts_dir}/add_cron_job.sh",
        ]
    )

    logger.info("Removing setup scripts from instance")
    subprocess.run(["ssh", "-i", pem_file, host, "rm", "-rf", host_scripts_dir])

    lambda_function_filename = os.path.join(scripts_dir, "lambda_start_ec2.py")
    project_prefix = project_name.replace("_", "-").replace(".", "-").lower()
    lambda_handler_name = "lambda_start_ec2.lambda_handler"
    lambda_role_name = (
        f"{project_prefix}-lambda-role" if role_name is None else role_name
    )

    # Extracting pipeline name from pipeline config file
    pipeline_name = os.path.splitext(os.path.basename(pipeline_config_file))[0]

    lambda_function_name = f"{project_prefix}-{pipeline_name}-lambda-scheduled"
    event_rule_name = f"{project_prefix}-{pipeline_name}-event-scheduled"

    iam_resource = boto_session.resource("iam")
    lambda_client = boto_session.client("lambda")
    eventbridge_client = boto_session.client("events")

    logger.info(
        f"Creating AWS Lambda function {lambda_function_name} from the "
        f"{lambda_handler_name} function in {lambda_function_filename}..."
    )

    deployment_package = create_lambda_deployment_package(
        lambda_function_filename
    )
    iam_role = create_iam_role_for_lambda(iam_resource, lambda_role_name)

    logger.info(
        f"Attempting to delete existing Lambda function {lambda_function_name}..."
    )
    try:
        delete_lambda_function(lambda_client, lambda_function_name)
        logger.info(f"Deleted existing Lambda function {lambda_function_name}")
    except Exception as e:
        logger.info(
            f"Failed to delete or Lambda function {lambda_function_name} not found."
        )
        logger.info(str(e))
        pass

    lambda_function_arn = exponential_retry(
        deploy_lambda_function,
        "InvalidParameterValueException",
        lambda_client,
        lambda_function_name,
        lambda_handler_name,
        iam_role,
        deployment_package,
    )

    logger.info(
        f"Attempting to delete existing Event Rule {event_rule_name}..."
    )
    try:
        delete_event_rule(
            eventbridge_client, event_rule_name, lambda_function_name
        )
        logger.info(f"Deleted existing Event Rule {event_rule_name}")
    except Exception as e:
        logger.info(
            f"Failed to delete or Event Rule {event_rule_name} not found."
        )
        logger.info(str(e))
        pass

    logger.info(
        f"Scheduling {lambda_function_name} to run with rate {event_schedule}..."
    )
    schedule_lambda_function(
        eventbridge_client,
        event_rule_name,
        event_schedule,
        lambda_client,
        lambda_function_name,
        lambda_function_arn,
    )

    logger.info("Removing setup scripts from local machine temporary directory")
    shutil.rmtree(scripts_dir)


def enable_or_disable_event_bridge_rule(
    credentials_file: str,
    project_name: str,
    pipeline: str,
    enabled: bool = True,
):
    """Enable or disable an event bridge rule.

    Args:
        credentials_file: Path to the AWS credentials file.
        project_name: Name of the project.
        enabled: True to enable the rule, False to disable the rule.
    """
    creds = read_yaml(credentials_file)
    boto_session = get_boto_session(creds)
    eventbridge_client = boto_session.client("events")

    pipeline_name = os.path.splitext(os.path.basename(pipeline))[0]
    project_prefix = project_name.replace("_", "-").replace(".", "-").lower()
    event_rule_name = f"{project_prefix}-{pipeline_name}-event-scheduled"
    update_event_rule(eventbridge_client, event_rule_name, enabled)


@click.command(epilog="""Disable event bridge rule for the project.""")
@click.option(
    "-c",
    "--credentials-file",
    default="credentials.yaml",
    help="Path to the AWS credentials file.",
)
@click.option(
    "-p",
    "--pipeline",
    required=True,
    type=click.Path(exists=True),
    help="Path to the pipeline configuration file.",
)
@click.option(
    "-pn",
    "--project-name",
    required=True,
    type=str,
    help="Path to the project repository.",
)
@verbosity_option()
@rudderlabs.raise_on_error
def disable_scheduling(credentials_file, pipeline, project_name):
    """Disable an event bridge rule for the project."""
    enable_or_disable_event_bridge_rule(
        credentials_file, project_name, pipeline, False
    )


@click.command(epilog="""Enable event bridge rule for the project.""")
@click.option(
    "-c",
    "--credentials-file",
    default="credentials.yaml",
    help="Path to the AWS credentials file.",
)
@click.option(
    "-p",
    "--pipeline",
    required=True,
    type=click.Path(exists=True),
    help="Path to the pipeline configuration file.",
)
@click.option(
    "-pn",
    "--project-name",
    required=True,
    type=str,
    help="Path to the project repository.",
)
@verbosity_option()
@rudderlabs.raise_on_error
def enable_scheduling(credentials_file, pipeline, project_name):
    """Enable an event bridge rule for the project."""
    enable_or_disable_event_bridge_rule(
        credentials_file, project_name, pipeline, True
    )


# the group command must be at the end of this file for plugins to work.
@with_plugins(iter_entry_points("rlabs.aws.pipeline.cli"))
@click.group(
    epilog="""Examples:

\b
    # Running pipeline
    ------------------

    $ rlabs aws pipeline run --pipeline-config-file pipeline.yaml --credentials-file credentials.yaml --repository-path /path/to/repository --instance-type ml.t3.xlarge --job-id my-job-id

    $ rlabs aws pipeline run -p pipeline.yaml -c credentials.yaml -r /path/to/repository -i local -j  Scheduling pipeline
    -----------------------

    $ rlabs aws pipeline schedule --instance-id <instance_id> --pem-file <path_to_pem_file> -u <aws_instance_user_name> --pipeline-config-file pipeline.yaml --credentials-file credentials.yaml --repository-path /path/to/repository --sagemaker-instance-type ml.t3.xlarge --event-schedule rate(1 hour) --role-name <aws_role_name> -vv

    $ rlabs aws pipeline schedule -id i-07c3cedbe6988ed49 -p ~/.ssh/instance.pem -u ubuntu -pc pipelines/sample_pipeline.yaml -c credentials_modified.yaml -es "rate(10 minutes)" -vv

    # Enabling and disabling pipeline event rule
    --------------------------------------------

    $ rlabs aws pipeline enable-scheduling --credentials-file credentials.yaml --repository-path /path/to/repository

    $ rlabs aws pipeline disable-scheduling --credentials-file credentials.yaml --repository-path /path/to/repository

"""
)
def pipeline():
    """Pipeline deployment scripts"""
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Amazon RedShift Connector."""

import urllib.parse
import pandas as pd
import pandas_redshift as pr

from sqlalchemy import create_engine
from sqlalchemy import orm as sa_orm

from ..log import get_logger
from .connector_base import ConnectorBase, register_connector

logger = get_logger(__name__)

@register_connector
class RedShiftConnector(ConnectorBase):
    def __init__(self, creds: dict, db_config: dict, **kwargs) -> None:
        super().__init__(creds, db_config, **kwargs)

        if "aws_config" not in kwargs:
            Exception("No aws_config found")

        self.aws_config = kwargs["aws_config"]
        encoded_password = urllib.parse.quote(creds["password"], safe="")
        connection_string = f"postgresql://{creds['user']}:{encoded_password}@{creds['host']}:{creds['port']}/{db_config['database']}"
        self.engine = create_engine(connection_string)

        Session = sa_orm.sessionmaker()
        Session.configure(bind=self.engine)
        self.connection = Session()
        if db_config.get("schema", None):
            self.connection.execute(f"SET search_path TO {db_config['schema']}")
            self.connection.commit()

    def write_to_table(
        self,
        df: pd.DataFrame,
        table_name: str,
        schema: str = None,
        if_exists: str = "append",
    ):
        table_name, schema = (
            table_name.split(".") if "." in table_name else (table_name, schema)
        )

        try:
            pr.connect_to_redshift(
                dbname=self.db_config["database"],
                host=self.creds["host"],
                port=self.creds["port"],
                user=self.creds["user"],
                password=self.creds["password"],
            )

            s3_bucket = self.creds.get("s3Bucket", None)
            s3_bucket = (
                s3_bucket if s3_bucket is not None else self.aws_config["s3Bucket"]
            )

            s3_sub_dir = self.creds.get("s3SubDirectory", None)
            s3_sub_dir = (
                s3_sub_dir
                if s3_sub_dir is not None
                else self.aws_config["s3SubDirectory"]
            )

            pr.connect_to_s3(
                aws_access_key_id=self.aws_config["access_key_id"],
                aws_secret_access_key=self.aws_config["access_key_secret"],
                bucket=s3_bucket,
                subdirectory=s3_sub_dir
                # As of release 1.1.1 you are able to specify an aws_session_token (if necessary):
                # aws_session_token = <aws_session_token>
            )

            # Write the DataFrame to S3 and then to redshift
            pr.pandas_to_redshift(
                data_frame=df,
                redshift_table_name=f"{schema}.{table_name}",
                append=if_exists == "append",
            )
        except Exception as e:
            logger.error(f"Error while writing to Redshift: {e}")

            #Check for non existing schema
            if "cannot copy into nonexistent table" in str(e).lower():
                self.create_table(df, table_name, schema)
                # Try again
                logger.info("Trying again")
                self.write_to_table(df, table_name, schema, if_exists)



#!/usr/bin/env python

import os

import click
import jinja2

from ..constants import SAGEMAKER_CONTAINER_PATH_MAIN
from ..log import get_logger, verbosity_option
from ..utils import render_template
from . import rudderlabs

logger = get_logger(__name__)

@click.command(
    epilog="""
Examples:

  1. Generates a new project for Bob:

     $ rlabs new -vv data-apps-leadscoring -t "Lead Scoring" -o ~/Projects

     $ rlabs new -vv <project> -t <title> -o <output_dir>
"""
)
@click.argument("project")
@click.option(
    "-t",
    "--title",
    show_default=True,
    default="New project",
    help="This entry defines the project title. "
    "The project title should be a few words only.  It will appear "
    "at the description of your project and as the title of your "
    "documentation",
)
@click.option(
    "-o",
    "--output-dir",
    help="Directory where to dump the new " "project - must not exist",
)
@verbosity_option()
@rudderlabs.raise_on_error
def new(project, title, output_dir):
    """Creates a folder structure for a new rudderlabs data apps project."""

    # the jinja context defines the substitutions to be performed
    context = dict(
        project=project,
        title=title,
        sagemaker_container_path=SAGEMAKER_CONTAINER_PATH_MAIN,
    )

    # copy the whole template structure and de-templatize the needed files
    if output_dir is None:
        output_dir = os.path.join(os.path.realpath(os.curdir), project)

    logger.info(
        "Creating structure for %s at directory %s", project, output_dir
    )

    project_dir = os.path.join(output_dir, project)
    if os.path.exists(project_dir):
        raise IOError(
            "The project directory %s already exists - cannot "
            "overwrite!" % project_dir
        )

    logger.info("mkdir %s", project_dir)
    os.makedirs(project_dir)

    # base jinja2 engine
    template_loader = jinja2.PackageLoader("rudderlabs.data.apps", os.path.join("templates","project"))
    env = jinja2.Environment(
        loader = template_loader,
        autoescape = jinja2.select_autoescape(["html", "xml"]),
    )

    #Render all files in the template directory
    for k in template_loader.list_templates():
        #ignore .pyc files
        if k.endswith(".pyc"):
            continue
        render_template(env, k, context, project_dir)

    logger.info(f"Creating base {project} structure in {output_dir}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""SnowFlake connector."""


import pandas as pd
import urllib.parse

from sqlalchemy import create_engine

from snowflake.connector.pandas_tools import write_pandas
from snowflake.connector import connect

from ..log import get_logger
from .connector_base import ConnectorBase, register_connector

logger = get_logger(__name__)

@register_connector
class SnowflakeConnector(ConnectorBase):
    def __init__(self, creds: dict, db_config: dict, **kwargs) -> None:
        super().__init__(creds, db_config, **kwargs)
        encoded_password = urllib.parse.quote(creds["password"], safe="") 
        url = f"snowflake://{creds['user']}:{encoded_password}@{creds['account_identifier']}"
        if "database" in db_config:
            url += f"/{db_config['database']}"
            if "schema" in db_config:
                url += f"/{db_config['schema']}"
                if "warehouse" in creds:
                    url += f"?warehouse={creds['warehouse']}"
                    if "role" in creds:
                        url += f"&role={creds['role']}"
        self.engine = create_engine(url)
        self.connection = self.engine.connect()
        self.creds = creds
        self.db_config = db_config

    def write_to_table(
        self,
        df: pd.DataFrame,
        table_name: str,
        schema: str = None,
        if_exists: str = "append",
    ):
        table_name, schema = (
            table_name.split(".") if "." in table_name else (table_name, schema)
        )
        print("Writing to table: {}.{}".format(schema, table_name))
        
        try:
            connection_params = {"user": self.creds["user"],
                                 "password": self.creds["password"],
                                 "account": self.creds["account_identifier"],
                                 "database": self.db_config["database"],
                                 "schema":schema}
            if "warehouse" in self.creds:
                connection_params["warehouse"] = self.creds["warehouse"]
            write_conn = connect(**connection_params)

            success, nchunks, nrows, _ = write_pandas(write_conn, df, table_name, quote_identifiers=False)

            if success:
                print(f"Successfully wrote {nrows} rows across {nchunks} chunks to table {table_name}")
            else:
                print("Write failed")
        except Exception as e:
            logger.error(f"Error while writing to Snowflake: {e}")

            #Check for non existing schema
            err_str = f"table '{table_name}' does not exist".lower()
            if err_str in str(e).lower():
                self.create_table(df, table_name, schema)
                # Try again
                logger.info("Trying again")
                self.write_to_table(df, table_name, schema, if_exists)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Shows how to use the AWS SDK for Python (Boto3) to create an AWS Lambda function,
invoke it, and delete it.
"""
import os
import io
import json
import random
import time
import zipfile
import boto3

from typing import Any, Callable
from botocore.exceptions import ClientError

from ..log import get_logger
logger = get_logger(__name__)

def exponential_retry(func, error_code, *func_args, **func_kwargs) -> Callable:
    """ 
    Retries the specified function with a simple exponential backoff algorithm.
    This is necessary when AWS is not yet ready to perform an action because all
    resources have not been fully deployed.

    Args:
        func: The function to retry.
        error_code: The error code to retry. Other errors are raised again.
        func_args: The positional arguments to pass to the function.
        func_kwargs: The keyword arguments to pass to the function.

    Returns:
        function: The return value of the retried function.
    """
    sleepy_time = 1
    func_return = None
    while sleepy_time < 33 and func_return is None:
        try:
            func_return = func(*func_args, **func_kwargs)
            logger.info("Ran %s, got %s.", func.__name__, func_return)
        except ClientError as error:
            if error.response['Error']['Code'] == error_code:
                print(f"Sleeping for {sleepy_time} to give AWS time to "
                      f"connect resources.")
                time.sleep(sleepy_time)
                sleepy_time = sleepy_time*2
            else:
                raise
    return func_return


def create_lambda_deployment_package(function_file_name) -> bytes:
    """ 
    Creates a Lambda deployment package in ZIP format in an in-memory buffer. This
    buffer can be passed directly to AWS Lambda when creating the function.

    Args:
        function_file_name: The name of the file that contains the Lambda handler function.

    Returns:
        bytes: The deployment package.
    """
    buffer = io.BytesIO()
    with zipfile.ZipFile(buffer, 'w') as zipped:
        zipped.write(function_file_name, os.path.basename(function_file_name))
    buffer.seek(0)
    return buffer.read()


def create_iam_role_for_lambda(iam_resource, iam_role_name) -> Any:
    """ 
    Creates an AWS Identity and Access Management (IAM) role that grants the
    AWS Lambda function basic permission to run. If a role with the specified
    name already exists, it is used for the demo.

    Args:
        iam_resource: The Boto3 IAM resource object.
        iam_role_name: The name of the role to create.

    Returns:
        Any: The newly created role.

    """
    lambda_assume_role_policy = {
        'Version': '2012-10-17',
        'Statement': [
            {
                'Effect': 'Allow',
                'Principal': {
                    'Service': 'lambda.amazonaws.com'
                },
                'Action': 'sts:AssumeRole'
            }
        ]
    }

    policy_arn = 'arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole'
    ec2_full_access_policy_arn = 'arn:aws:iam::aws:policy/AmazonEC2FullAccess'

    try:
        role = iam_resource.create_role(
            RoleName=iam_role_name,
            AssumeRolePolicyDocument=json.dumps(lambda_assume_role_policy))
        iam_resource.meta.client.get_waiter('role_exists').wait(RoleName=iam_role_name)
        logger.info("Created role %s.", role.name)

        role.attach_policy(PolicyArn=policy_arn)
        logger.info("Attached basic execution policy to role %s.", role.name)

        role.attach_policy(PolicyArn=ec2_full_access_policy_arn)
        logger.info("Attached EC2 full access policy to role %s.", role.name)

    except ClientError as error:
        if error.response['Error']['Code'] == 'EntityAlreadyExists':
            role = iam_resource.Role(iam_role_name)
            logger.warning("The role %s already exists. Using it.", iam_role_name)
        else:
            logger.exception(
                "Couldn't create role %s or attach policy %s, %s.",
                iam_role_name, policy_arn, ec2_full_access_policy_arn)
            raise

    return role


def deploy_lambda_function(lambda_client, function_name, handler_name, iam_role, deployment_package) -> Any:
    """ 
    Deploys the AWS Lambda function.

    Args:
        lambda_client: The Boto3 AWS Lambda client object.
        function_name: The name of the AWS Lambda function.
        handler_name: The fully qualified name of the handler function. This
                      must include the file name and the function name.
        iam_role: The IAM role to use for the function.
        deployment_package: The deployment package that contains the function
                               code in ZIP format.

    Returns:
        Any: The Amazon Resource Name (ARN) of the newly created function.

    """
    try:
        response = lambda_client.create_function(
            FunctionName=function_name,
            Description="AWS Lambda demo",
            Runtime='python3.8',
            Role=iam_role.arn,
            Handler=handler_name,
            Code={'ZipFile': deployment_package},
            Publish=True)
        function_arn = response['FunctionArn']
        logger.info("Created function '%s' with ARN: '%s'.",
                    function_name, response['FunctionArn'])
    except ClientError as error:
        logger.info(error)
        logger.exception("Couldn't create function %s.", function_name)
        raise
    else:
        return function_arn


def delete_lambda_function(lambda_client, function_name) -> None:
    """ 
    Deletes an AWS Lambda function.

    Args:
        lambda_client: The Boto3 AWS Lambda client object.
        function_name: The name of the function to delete.

    Returns:
        None:  

    """
    try:
        lambda_client.delete_function(FunctionName=function_name)
    except ClientError:
        logger.exception("Couldn't delete function %s.", function_name)
        raise

def invoke_lambda_function(lambda_client, function_name, function_params) -> dict:
    """ 
    Invokes an AWS Lambda function.

    Args:
        lambda_client: The Boto3 AWS Lambda client object.
        function_name: The name of the function to invoke.
        function_params: The parameters of the function as a dict. This dict
                         is serialized to JSON before it is sent to AWS Lambda.

    Returns:
        dict: The response from the function invocation.

    """
    try:
        response = lambda_client.invoke(
            FunctionName=function_name,
            Payload=json.dumps(function_params).encode())
        logger.info("Invoked function %s.", function_name)
    except ClientError:
        logger.exception("Couldn't invoke function %s.", function_name)
        raise
    return response
